from .marisol import Area, BatesOverlay, Document, Marisol, OutsideBoundariesError, Page, Redaction, RedactionStyle, StaticOverlay  # noqa: F401 # pylint: disable=unused-import

__author__ = "Kevin Schellenberg"
__email__ = "wikkiewikkie@gmail.com"
__license__ = "MIT"
__version__ = "0.3.0"
